@@ -1,7 +1,7 @@
 /* sna.c: Routines for handling .sna snapshots
    Copyright (c) 2001-2002 Philip Kendall
 
-   $Id: sna.c,v 1.17 2003/11/27 21:45:25 pak21 Exp $
+   $Id: sna.c,v 1.23 2004/06/02 13:38:10 pak21 Exp $
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -291,6 +291,20 @@ libspectrum_sna_write( libspectrum_byte **buffer, size_t *length,
      etc which are not stored in .sna format */
   *out_flags = LIBSPECTRUM_FLAG_SNAPSHOT_MINOR_INFO_LOSS;
 
+  /* We don't save the ZXATASP or ZXCF at all */
+  if( libspectrum_snap_zxatasp_active( snap ) )
+    *out_flags |= LIBSPECTRUM_FLAG_SNAPSHOT_MAJOR_INFO_LOSS;
+  if( libspectrum_snap_zxcf_active( snap ) )
+    *out_flags |= LIBSPECTRUM_FLAG_SNAPSHOT_MAJOR_INFO_LOSS;
+
+  /* We don't save the Interface II ROM at all */
+  if( libspectrum_snap_interface2_active( snap ) )
+    *out_flags |= LIBSPECTRUM_FLAG_SNAPSHOT_MAJOR_INFO_LOSS;
+
+  /* We don't save the Timex Dock at all */
+  if( libspectrum_snap_dock_active( snap ) )
+    *out_flags |= LIBSPECTRUM_FLAG_SNAPSHOT_MAJOR_INFO_LOSS;
+
   ptr = *buffer;
 
   error = write_header( buffer, &ptr, length, &sp, snap );
@@ -309,7 +323,10 @@ libspectrum_sna_write( libspectrum_byte **buffer, size_t *length,
     
   case LIBSPECTRUM_MACHINE_PLUS2A:
   case LIBSPECTRUM_MACHINE_PLUS3:
+  case LIBSPECTRUM_MACHINE_PLUS3E:
   case LIBSPECTRUM_MACHINE_PENT:
+  case LIBSPECTRUM_MACHINE_SCORP:
+  case LIBSPECTRUM_MACHINE_SE:
     *out_flags |= LIBSPECTRUM_FLAG_SNAPSHOT_MAJOR_INFO_LOSS;
     /* Fall through */
   case LIBSPECTRUM_MACHINE_PLUS2:
@@ -321,13 +338,6 @@ libspectrum_sna_write( libspectrum_byte **buffer, size_t *length,
     libspectrum_print_error( LIBSPECTRUM_ERROR_LOGIC,
 			     "Emulated machine type is set to 'unknown'!" );
     return LIBSPECTRUM_ERROR_LOGIC;
-
-  default:
-    /* Should never happen */
-    libspectrum_print_error( LIBSPECTRUM_ERROR_LOGIC,
-			     "libspectrum_sna_write: unknown machine type %d",
-			     libspectrum_snap_machine( snap ) );
-    return LIBSPECTRUM_ERROR_LOGIC;
   }
 
   if( error ) return error;
