@@ -1,7 +1,7 @@
 /* tape.c: Routines for handling tape files
    Copyright (c) 2001-2003 Philip Kendall, Darren Salt
 
-   $Id: tape.c,v 1.63 2003/11/03 12:09:01 pak21 Exp $
+   $Id: tape.c,v 1.64 2004/04/07 10:06:41 pak21 Exp $
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -214,6 +214,38 @@ libspectrum_tape_read( libspectrum_tape *tape, const libspectrum_byte *buffer,
   return error;
 }
 
+libspectrum_error
+libspectrum_tape_write( libspectrum_byte **buffer, size_t *length,
+			libspectrum_tape *tape, libspectrum_id_t type )
+{
+  libspectrum_class_t class;
+  libspectrum_error error;
+
+  error = libspectrum_identify_class( &class, type );
+  if( error ) return error;
+
+  if( class != LIBSPECTRUM_CLASS_TAPE ) {
+    libspectrum_print_error( LIBSPECTRUM_ERROR_INVALID,
+			     "libspectrum_tape_write: not a tape format" );
+    return LIBSPECTRUM_ERROR_INVALID;
+  }
+
+  switch( type ) {
+
+  case LIBSPECTRUM_ID_TAPE_TAP:
+    return libspectrum_tap_write( buffer, length, tape );
+
+  case LIBSPECTRUM_ID_TAPE_TZX:
+    return libspectrum_tzx_write( buffer, length, tape );
+
+  default:
+    libspectrum_print_error( LIBSPECTRUM_ERROR_UNKNOWN,
+			     "libspectrum_tape_write: format not supported" );
+    return LIBSPECTRUM_ERROR_UNKNOWN;
+
+  }
+}
+
 /* Does this tape structure actually contain a tape? */
 int
 libspectrum_tape_present( const libspectrum_tape *tape )
