@@ -1,7 +1,7 @@
 /* z80.c: Routines for handling .z80 snapshots
-   Copyright (c) 2001-2002 Philip Kendall, Darren Salt
+   Copyright (c) 2001-2004 Philip Kendall, Darren Salt
 
-   $Id: z80.c,v 1.63 2003/11/07 14:06:19 pak21 Exp $
+   $Id: z80.c,v 1.75 2004/07/15 20:02:20 pak21 Exp $
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -44,6 +44,41 @@ static const int LIBSPECTRUM_Z80_HEADER_LENGTH = 30;
 /* Length of xzx's extensions */
 #define LIBSPECTRUM_Z80_V3X_LENGTH 55
 
+/* The constants used for each machine type */
+enum {
+
+  /* v2 constants */
+  Z80_MACHINE_48_V2 = 0,
+  Z80_MACHINE_48_IF1_V2 = 1,
+  Z80_MACHINE_48_SAMRAM_V2 = 2,
+  Z80_MACHINE_128_V2 = 3,
+  Z80_MACHINE_128_IF1_V2 = 4,
+
+  /* v3 constants */
+  Z80_MACHINE_48 = 0,
+  Z80_MACHINE_48_IF1 = 1,
+  Z80_MACHINE_48_SAMRAM = 2,
+  Z80_MACHINE_48_MGT = 3,
+  Z80_MACHINE_128 = 4,
+  Z80_MACHINE_128_IF1 = 5,
+  Z80_MACHINE_128_MGT = 6,
+
+  /* Extensions */
+  Z80_MACHINE_PLUS3 = 7,
+  Z80_MACHINE_PLUS3_XZX_ERROR = 8,
+  Z80_MACHINE_PENTAGON = 9,
+  Z80_MACHINE_SCORPION = 10,
+  Z80_MACHINE_PLUS2 = 12,
+  Z80_MACHINE_PLUS2A = 13,
+  Z80_MACHINE_TC2048 = 14,
+  Z80_MACHINE_TC2068 = 15,
+  Z80_MACHINE_TS2068 = 128,
+
+  /* The first extension ID; anything here or greater applies to both
+     v2 and v3 files */
+  Z80_MACHINE_FIRST_EXTENSION = Z80_MACHINE_PLUS3,
+};
+
 /* The signature used to designate the .slt extensions */
 static libspectrum_byte slt_signature[] = "\0\0\0SLT";
 static size_t slt_signature_length = 6;
@@ -52,6 +87,14 @@ static libspectrum_error
 read_header( const libspectrum_byte *buffer, libspectrum_snap *snap,
 	     const libspectrum_byte **data, int *version, int *compressed );
 static libspectrum_error
+get_machine_type( libspectrum_snap *snap, libspectrum_byte type, int version );
+static libspectrum_error
+get_machine_type_v2( libspectrum_snap *snap, libspectrum_byte type );
+static libspectrum_error
+get_machine_type_v3( libspectrum_snap *snap, libspectrum_byte type );
+static libspectrum_error
+get_machine_type_extension( libspectrum_snap *snap, libspectrum_byte type );
+static libspectrum_error
 read_blocks( const libspectrum_byte *buffer, size_t buffer_length,
 	     libspectrum_snap *snap, int version, int compressed );
 static libspectrum_error
@@ -73,13 +116,13 @@ read_v2_block( const libspectrum_byte *buffer, libspectrum_byte **block,
 
 static libspectrum_error
 write_header( libspectrum_byte **buffer, libspectrum_byte **ptr,
-	      size_t *length, libspectrum_snap *snap );
+	      size_t *length, int *flags, libspectrum_snap *snap );
 static libspectrum_error
 write_base_header( libspectrum_byte **buffer, libspectrum_byte **ptr,
 		   size_t *length, libspectrum_snap *snap );
 static libspectrum_error
 write_extended_header( libspectrum_byte **buffer, libspectrum_byte **ptr,
-		       size_t *length, libspectrum_snap *snap );
+		       size_t *length, int *flags, libspectrum_snap *snap );
 static libspectrum_error
 write_pages( libspectrum_byte **buffer, libspectrum_byte **ptr,
 	     size_t *length, libspectrum_snap *snap, int compress );
@@ -139,6 +182,7 @@ read_header( const libspectrum_byte *buffer, libspectrum_snap *snap,
   const libspectrum_byte *header = buffer;
   int capabilities;
   size_t i;
+  libspectrum_error error;
 
   libspectrum_snap_set_a  ( snap, header[ 0] );
   libspectrum_snap_set_f  ( snap, header[ 1] );
@@ -195,64 +239,10 @@ read_header( const libspectrum_byte *buffer, libspectrum_snap *snap,
 
     libspectrum_snap_set_pc( snap, extra_header[0] + extra_header[1] * 0x100 );
 
-    switch( *version ) {
+    error = get_machine_type( snap, extra_header[2], *version );
+    if( error ) return error;
 
-    case 2:
-
-      switch( extra_header[2] ) {
-      case 0: case 1: case 2:
-	libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_48 ); break;
-      case 3: case 4:
-	libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_128 ); break;
-      case 14:
-	libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_TC2048 );
-	break;
-      case 15:
-      case 128: /* Load TS2068 snaps as TC2068 for now */
-	libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_TC2068 );
-	break;
-      default:
-        libspectrum_print_error(
-          LIBSPECTRUM_ERROR_UNKNOWN,
-          "libspectrum_read_z80_header: unknown v2 machine type %d",
-	  extra_header[2]
-        );
-	return LIBSPECTRUM_ERROR_UNKNOWN;
-      }
-
-      break;
-
-    case 3:
-
-      switch( extra_header[2] ) {
-      case 0: case 1: case 2: case 3:
-	libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_48 ); break;
-      case 4: case 5: case 6:
-	libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_128 ); break;
-      case 7: case 8: /* 8 mistakenly written by some versions of xzx */
-	libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_PLUS3 ); break;
-      case 9:
-	libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_PENT ); break;
-      case 12:
-	libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_PLUS2 ); break;
-      case 13:
-	libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_PLUS2A );
-	break;
-      case 14:
-	libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_TC2048 );
-	break;
-      case 15:
-      case 128: /* Load TS2068 snaps as TC2068 for now */
-	libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_TC2068 );
-	break;
-      default:
-        libspectrum_print_error(
-          LIBSPECTRUM_ERROR_UNKNOWN,
-          "libspectrum_read_z80_header: unknown v3 machine type %d",
-	  extra_header[2]
-        );
-	return LIBSPECTRUM_ERROR_UNKNOWN;
-      }
+    if( *version >= 3 ) {
 
       quarter_tstates =
 	libspectrum_timings_tstates_per_frame(
@@ -270,16 +260,6 @@ read_header( const libspectrum_byte *buffer, libspectrum_snap *snap,
       /* Stop broken files from causing trouble... */
       if( libspectrum_snap_tstates( snap ) >= quarter_tstates * 4 )
 	libspectrum_snap_set_tstates( snap, 0 );
-      
-      break;
-
-    default:
-      libspectrum_print_error(
-        LIBSPECTRUM_ERROR_LOGIC,
-        "libspectrum_read_z80_header: unknown snap version %d", *version
-      );
-      return LIBSPECTRUM_ERROR_LOGIC;
-
     }
 
     /* I don't like this any more than you do but here is support for the
@@ -337,9 +317,12 @@ read_header( const libspectrum_byte *buffer, libspectrum_snap *snap,
       }
     }
 
-    if( ( capabilities & LIBSPECTRUM_MACHINE_CAPABILITY_PLUS3_MEMORY ) &&
-	( extra_length == LIBSPECTRUM_Z80_V3X_LENGTH )                    )
-      libspectrum_snap_set_out_plus3_memoryport( snap, extra_header[54] );
+    if( extra_length == LIBSPECTRUM_Z80_V3X_LENGTH ) {
+      if( ( capabilities & LIBSPECTRUM_MACHINE_CAPABILITY_PLUS3_MEMORY ) ||
+	  ( capabilities & LIBSPECTRUM_MACHINE_CAPABILITY_SCORP_MEMORY )    ) {
+	libspectrum_snap_set_out_plus3_memoryport( snap, extra_header[54] );
+      }
+    }
 
     (*data) = buffer + LIBSPECTRUM_Z80_HEADER_LENGTH + 2 + extra_length;
 
@@ -359,6 +342,115 @@ read_header( const libspectrum_byte *buffer, libspectrum_snap *snap,
 
 }
 
+static libspectrum_error
+get_machine_type( libspectrum_snap *snap, libspectrum_byte type, int version )
+{
+  libspectrum_error error;
+
+  if( type < Z80_MACHINE_FIRST_EXTENSION ) {
+
+    switch( version ) {
+
+    case 2:
+      error = get_machine_type_v2( snap, type ); if( error ) return error;
+      break;
+
+    case 3:
+      error = get_machine_type_v3( snap, type ); if( error ) return error;
+      break;
+
+    default:
+      libspectrum_print_error( LIBSPECTRUM_ERROR_LOGIC,
+			       "%s:get_machine_type: unknown version %d",
+			       __FILE__, version );
+      return LIBSPECTRUM_ERROR_LOGIC;
+    }
+
+  } else {
+    error = get_machine_type_extension( snap, type ); if( error ) return error;
+  }
+
+  return LIBSPECTRUM_ERROR_NONE;
+}
+
+static libspectrum_error
+get_machine_type_v2( libspectrum_snap *snap, libspectrum_byte type )
+{
+  switch( type ) {
+
+  case Z80_MACHINE_48_V2:
+  case Z80_MACHINE_48_IF1_V2:
+  case Z80_MACHINE_48_SAMRAM_V2:
+    libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_48 ); break;
+  case Z80_MACHINE_128_V2:
+  case Z80_MACHINE_128_IF1_V2:
+    libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_128 ); break;
+  default:
+    libspectrum_print_error( LIBSPECTRUM_ERROR_UNKNOWN,
+			     "%s:get_machine_type: unknown v2 machine type %d",
+			     __FILE__, type );
+    return LIBSPECTRUM_ERROR_UNKNOWN;
+  }
+
+  return LIBSPECTRUM_ERROR_NONE;
+}
+
+static libspectrum_error
+get_machine_type_v3( libspectrum_snap *snap, libspectrum_byte type )
+{
+  switch( type ) {
+    
+  case Z80_MACHINE_48:
+  case Z80_MACHINE_48_IF1:
+  case Z80_MACHINE_48_SAMRAM:
+  case Z80_MACHINE_48_MGT:
+    libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_48 ); break;
+  case Z80_MACHINE_128:
+  case Z80_MACHINE_128_IF1:
+  case Z80_MACHINE_128_MGT:
+    libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_128 ); break;
+  default:
+    libspectrum_print_error( LIBSPECTRUM_ERROR_UNKNOWN,
+			     "%s:get_machine_type: unknown v3 machine type %d",
+			     __FILE__, type );
+    return LIBSPECTRUM_ERROR_UNKNOWN;
+  }
+
+  return LIBSPECTRUM_ERROR_NONE;
+}
+
+static libspectrum_error
+get_machine_type_extension( libspectrum_snap *snap, libspectrum_byte type )
+{
+  switch( type ) {
+
+  case Z80_MACHINE_PLUS3:
+  case Z80_MACHINE_PLUS3_XZX_ERROR:
+    libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_PLUS3 ); break;
+  case Z80_MACHINE_PENTAGON:
+    libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_PENT ); break;
+  case Z80_MACHINE_SCORPION:
+    libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_SCORP ); break;
+  case Z80_MACHINE_PLUS2:
+    libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_PLUS2 ); break;
+  case Z80_MACHINE_PLUS2A:
+    libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_PLUS2A ); break;
+  case Z80_MACHINE_TC2048:
+    libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_TC2048 ); break;
+  case Z80_MACHINE_TC2068:
+  case Z80_MACHINE_TS2068: /* Load TS2068 snaps as TC2068 for now */
+    libspectrum_snap_set_machine( snap, LIBSPECTRUM_MACHINE_TC2068 ); break;
+  default:
+    libspectrum_print_error(
+      LIBSPECTRUM_ERROR_UNKNOWN,
+      "%s:get_machine_type: unknown extension machine type %d", __FILE__, type
+    );
+    return LIBSPECTRUM_ERROR_UNKNOWN;
+  }
+
+  return LIBSPECTRUM_ERROR_NONE;
+}
+
 static libspectrum_error
 read_blocks( const libspectrum_byte *buffer, size_t buffer_length,
 	     libspectrum_snap *snap, int version, int compressed )
@@ -581,14 +673,22 @@ read_block( const libspectrum_byte *buffer, libspectrum_snap *snap,
 			   end );
     if( error != LIBSPECTRUM_ERROR_NONE ) return error;
 
-    if( page <= 0 || page > 11 ) {
+    if( page <= 0 || page > 18 ) {
       libspectrum_print_error( LIBSPECTRUM_ERROR_UNKNOWN,
 			       "read_block: unknown page %d", page );
       return LIBSPECTRUM_ERROR_UNKNOWN;
     }
 
     /* If it's a ROM page, just throw it away */
-    if( page < 3 || page > 10 ) {
+    if( page < 3 ) {
+      free( uncompressed );
+      return LIBSPECTRUM_ERROR_NONE;
+    }
+
+    /* Page 11 is the Multiface ROM unless we're emulating something
+       Scorpion-like */
+    if( page == 11 &&
+	!( capabilities & LIBSPECTRUM_MACHINE_CAPABILITY_SCORP_MEMORY ) ) {
       free( uncompressed );
       return LIBSPECTRUM_ERROR_NONE;
     }
@@ -818,7 +918,22 @@ libspectrum_z80_write2( libspectrum_byte **buffer, size_t *length,
   if( libspectrum_snap_halted( snap ) )
     *out_flags |= LIBSPECTRUM_FLAG_SNAPSHOT_MINOR_INFO_LOSS;
 
-  error = write_header( buffer, &ptr, length, snap );
+  /* .z80 format doesn't store the ZXATASP or ZXCF info */
+  if( libspectrum_snap_zxatasp_active( snap ) )
+    *out_flags |= LIBSPECTRUM_FLAG_SNAPSHOT_MAJOR_INFO_LOSS;
+  if( libspectrum_snap_zxcf_active( snap ) )
+    *out_flags |= LIBSPECTRUM_FLAG_SNAPSHOT_MAJOR_INFO_LOSS;
+
+  /* .z80 format may be able to store an IF2 ROM instead of the 48K ROM but
+     not for now */
+  if( libspectrum_snap_interface2_active( snap ) )
+    *out_flags |= LIBSPECTRUM_FLAG_SNAPSHOT_MAJOR_INFO_LOSS;
+
+  /* .z80 format doesn't store the Timex Dock cart */
+  if( libspectrum_snap_dock_active( snap ) )
+    *out_flags |= LIBSPECTRUM_FLAG_SNAPSHOT_MAJOR_INFO_LOSS;
+
+  error = write_header( buffer, &ptr, length, out_flags, snap );
   if( error != LIBSPECTRUM_ERROR_NONE ) return error;
 
   error = write_pages(
@@ -834,14 +949,14 @@ libspectrum_z80_write2( libspectrum_byte **buffer, size_t *length,
 
 static libspectrum_error
 write_header( libspectrum_byte **buffer, libspectrum_byte **ptr,
-	      size_t *length, libspectrum_snap *snap )
+	      size_t *length, int *flags, libspectrum_snap *snap )
 {
   libspectrum_error error;
 
   error = write_base_header( buffer, ptr, length, snap );
   if( error != LIBSPECTRUM_ERROR_NONE ) return error;
 
-  error = write_extended_header( buffer, ptr, length, snap );
+  error = write_extended_header( buffer, ptr, length, flags, snap );
   if( error != LIBSPECTRUM_ERROR_NONE ) return error;
 
   return LIBSPECTRUM_ERROR_NONE;
@@ -889,9 +1004,9 @@ write_base_header( libspectrum_byte **buffer, libspectrum_byte **ptr,
 
 static libspectrum_error
 write_extended_header( libspectrum_byte **buffer, libspectrum_byte **ptr,
-		       size_t *length, libspectrum_snap *snap )
+		       size_t *length, int *flags, libspectrum_snap *snap )
 {
-  int i; libspectrum_error error;
+  int i, bottom_16kb_ram; libspectrum_error error;
 
   libspectrum_dword quarter_states;
 
@@ -910,25 +1025,34 @@ write_extended_header( libspectrum_byte **buffer, libspectrum_byte **ptr,
   switch( libspectrum_snap_machine( snap ) ) {
   case LIBSPECTRUM_MACHINE_16:
   case LIBSPECTRUM_MACHINE_48:
-    *(*ptr)++ = 0; break;
+    *(*ptr)++ = Z80_MACHINE_48; break;
+  case LIBSPECTRUM_MACHINE_SE:
+    *flags |= LIBSPECTRUM_FLAG_SNAPSHOT_MAJOR_INFO_LOSS;
+    /* fall through */
   case LIBSPECTRUM_MACHINE_128:
-    *(*ptr)++ = 4; break;
+    *(*ptr)++ = Z80_MACHINE_128; break;
+  case LIBSPECTRUM_MACHINE_PLUS3E:
+    *flags |= LIBSPECTRUM_FLAG_SNAPSHOT_MINOR_INFO_LOSS;
+    /* fall through */
   case LIBSPECTRUM_MACHINE_PLUS3:
-    *(*ptr)++ = 7; break;
+    *(*ptr)++ = Z80_MACHINE_PLUS3; break;
   case LIBSPECTRUM_MACHINE_PENT:
-    *(*ptr)++ = 9; break;
+    *(*ptr)++ = Z80_MACHINE_PENTAGON; break;
+  case LIBSPECTRUM_MACHINE_SCORP:
+    *(*ptr)++ = Z80_MACHINE_SCORPION; break;
   case LIBSPECTRUM_MACHINE_PLUS2:
-    *(*ptr)++ = 12; break;
+    *(*ptr)++ = Z80_MACHINE_PLUS2; break;
   case LIBSPECTRUM_MACHINE_PLUS2A:
-    *(*ptr)++ = 13; break;
+    *(*ptr)++ = Z80_MACHINE_PLUS2A; break;
   case LIBSPECTRUM_MACHINE_TC2048:
-    *(*ptr)++ = 14; break;
+    *(*ptr)++ = Z80_MACHINE_TC2048; break;
   case LIBSPECTRUM_MACHINE_TC2068:
-    *(*ptr)++ = 15; break;
-  default:
+    *(*ptr)++ = Z80_MACHINE_TC2068; break;
+
+  case LIBSPECTRUM_MACHINE_UNKNOWN:
     libspectrum_print_error( LIBSPECTRUM_ERROR_UNKNOWN,
-			     "write_extended_header: unknown machine type %d",
-			     libspectrum_snap_machine( snap ) );
+			     "%s:write_extended_header: machine type unknown",
+			     __FILE__ );
     return LIBSPECTRUM_ERROR_UNKNOWN;
   }
 
@@ -974,10 +1098,22 @@ write_extended_header( libspectrum_byte **buffer, libspectrum_byte **ptr,
   /* Spectator, MGT and Multiface disabled */
   *(*ptr)++ = '\0'; *(*ptr)++ = '\0'; *(*ptr)++ = '\0';
 
-  /* Is 0x0000 to 0x7fff RAM? Currently iff we're in a +3 special
-     configuration */
-  if( ( capabilities & LIBSPECTRUM_MACHINE_CAPABILITY_PLUS3_MEMORY ) &&
-      ( libspectrum_snap_out_plus3_memoryport( snap ) & 0x01                          )    ) {
+  /* Is 0x0000 to 0x3fff RAM? True if we're in a +3 64Kb RAM
+     configuration, or a Scorpion configuration with page 0 mapped in */
+  bottom_16kb_ram = 0;
+
+  if( capabilities & LIBSPECTRUM_MACHINE_CAPABILITY_PLUS3_MEMORY ) {
+
+    if( libspectrum_snap_out_plus3_memoryport( snap ) & 0x01 )
+      bottom_16kb_ram = 1;
+
+  } else if( capabilities & LIBSPECTRUM_MACHINE_CAPABILITY_SCORP_MEMORY ) {
+
+    if( libspectrum_snap_out_plus3_memoryport( snap ) & 0x01 )
+      bottom_16kb_ram = 1;
+  }
+
+  if( bottom_16kb_ram ) {
     *(*ptr)++ = 0xff; *(*ptr)++ = 0xff;
   } else {
     *(*ptr)++ = 0x00; *(*ptr)++ = 0x00;
@@ -1023,6 +1159,16 @@ write_pages( libspectrum_byte **buffer, libspectrum_byte **ptr, size_t *length,
       }
     }
 
+    if( capabilities & LIBSPECTRUM_MACHINE_CAPABILITY_SCORP_MEMORY ) {
+      for( i = 8; i < 16; i++ ) {
+        if( libspectrum_snap_pages( snap, i ) ) {
+          error = write_page( buffer, ptr, length, i + 3,
+                              libspectrum_snap_pages( snap, i ), compress );
+          if( error != LIBSPECTRUM_ERROR_NONE ) return error;
+        }
+      }
+    }
+
   }
 
   /* Do we want to write .slt data? Definitely if we've got a loading
